/*
 *	ACCES I/O PCI-WDG-CSM driver
 *
 *	(c) Copyright 1998-2012 Jimi Damon<jdamon@accesio.com>,
 *						All Rights Reserved.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Neither Jimi Damon nor Acces I/O admit liability nor provide
 *	warranty for any of this software. This material is provided
 *	"AS-IS" and at no charge.
 *
 *	(c) Copyright 2012   Jimi Damon <jdamon@accesio.com>
 *
 *	Release 0.1
 *
 *
 */

#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/io.h>
#include <linux/uaccess.h>

#include <asm/system.h>

#define WDT_IS_PCI
#include "wd501p.h"
#include "aio_wdt.h"


#define PFX "aio_wdt: "

/*
 * Until Access I/O gets their application for a PCI vendor ID approved,
 * I don't think that it's appropriate to move these constants into the
 * regular pci_ids.h file. -- JPN 2000/01/18
 */

/* Types of Boards
 *
 *
 * WDG-CSM Provides
 * a. Two isolated digital inputs, current limited.
 * b. Two isolated digital outputs, current limited.
 * c. Two non-isolated digital outputs in parallel to isolated digital outputs.
 * d. Reset pushbutton input, TTL, active LOW, buffered.
 * e. Fused external 5VDC power for external LED.
 *
 *
 * P104-WDG-CSMA
 * f. Two undedicated analog inputs, 0 to +5V, eight-bit conversion.
 * g. One remote reset isolated input, current limit
 * h. Fan drive, PWM, open collector.
 * i. Additional +5 volts through a resettable fuse (300mA at room temperature).
 * j. +12 volts through a resettable fuse (300mA at room temperature).
 */

#ifndef PCI_VENDOR_ID_ACCESSIO
#define PCI_VENDOR_ID_ACCESSIO 0x494f
#endif
#ifndef PCI_DEVICE_ID_WDG_CSM
#define PCI_DEVICE_ID_WDG_CSM 0x22c0
#endif

/* We can only use 1 card due to the /dev/watchdog restriction */
static int dev_count;

static unsigned long open_lock;
static DEFINE_SPINLOCK(aio_wdt_lock);
static char expect_close;

static resource_size_t io;
static int irq;

/* Default timeout */
#define WD_TIMO 60			/* Default heartbeat = 60 seconds */

static int heartbeat = WD_TIMO;
static int wd_heartbeat;
module_param(heartbeat, int, 0);
MODULE_PARM_DESC(heartbeat,
		"Watchdog heartbeat in seconds. (0<heartbeat<65536, default="
				__MODULE_STRING(WD_TIMO) ")");

static int nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout,
		 "Watchdog cannot be stopped once started (default="
				__MODULE_STRING(WATCHDOG_NOWAYOUT) ")");

/* Support for the Fan Tachometer on the PCI-WDT501 */
/* static int tachometer; */

static int tachometer;
module_param(tachometer, int, 0);
MODULE_PARM_DESC(tachometer,
		"PCI-WDT501 Fan Tachometer support (0=disable, default=0)");


static int type = 0;
module_param(type, int, 0);
MODULE_PARM_DESC(type,"ACCES I/O PCI_WDT type (CSM or CSMA , default=0)");

/*
 * Analog voltage support
 */ 


/* 
 * Computer Power monitor 
 */


/* 
 * Security Light Sensor 
 */


/*
 * Opto-Isolated Outputs
 */


static void aio_wdt_ctr_mode(char cntr, char mode)
{
  int ctrl;
  ctrl = (cntr << 6) | 0x30 | (mode << 1);
  outb(ctrl,WDT_CR );
}

static void aio_wdt_ctr_load(int c,int val)
{
  outb( val & 0x00FF      ,  WDT_COUNT0 + c );
  outb( (val>>8) & 0x00FF ,  WDT_COUNT0 + c );
}

/**
 *	aio_wdt_start:
 *
 *	Start the watchdog driver.
 */
static int aio_wdt_start(void)
{
	unsigned long flags;

	spin_lock_irqsave(&aio_wdt_lock, flags);
#ifdef AIO_WDG_DEBUG
	printk(KERN_CRIT PFX "starting wdog , status %u\n", (unsigned)WDT_DC );
#endif

	/*
	 * "pet" the watchdog, as Access says.
	 * This resets the clock outputs.
	 */
        udelay(8);
	inb(WDT_DC);			/* Disable watchdog */

        aio_wdt_ctr_mode(  0, 3 );
        aio_wdt_ctr_load(  0, 20833 );
        aio_wdt_ctr_mode(  1, 2 );
        aio_wdt_ctr_load(  1, wd_heartbeat );
        aio_wdt_ctr_mode(  2, 1 );

	outb(0, WDT_DC);		/* Enable watchdog */
	udelay(8);

	spin_unlock_irqrestore(&aio_wdt_lock, flags);
	return 0;
}

/**
 *	aio_wdt_stop:
 *
 *	Stop the watchdog driver.
 */
static int aio_wdt_stop(void)
{
	unsigned long flags;
	/* Turn the card off */
	spin_lock_irqsave(&aio_wdt_lock, flags);
	inb(WDT_DC);
	udelay(8);
	/* aio_wdt_ctr_load(2, 0);		/\* 0 length reset pulses now *\/ */
	spin_unlock_irqrestore(&aio_wdt_lock, flags);
#ifdef AIO_WDG_DEBUG
        printk(KERN_ERR PFX "Stopping counter temporarily\n");
#endif
	return 0;
}

/**
 *	aio_wdt_ping:
 *
 *	Reload counter one with the watchdog heartbeat. We don't bother
 *	reloading the cascade counter.
 */

static int aio_wdt_ping(void)
{
	unsigned long flags;

	spin_lock_irqsave(&aio_wdt_lock, flags);
	/* Write a watchdog value */
	inb(WDT_DC);			/* Disable watchdog */
	udelay(8);
 
        aio_wdt_ctr_mode( 1, 2);
        aio_wdt_ctr_load( 1,6000);

	outb(0, WDT_DC);		/* Enable watchdog */
	udelay(8);
	spin_unlock_irqrestore(&aio_wdt_lock, flags);
	return 0;
}

/**
 *	aio_wdt_set_heartbeat:
 *	@t:		the new heartbeat value that needs to be set.
 *
 *	Set a new heartbeat value for the watchdog device. If the heartbeat
 *	value is incorrect we keep the old value and return -EINVAL.
 *	If successful we return 0.
 */
static int aio_wdt_set_heartbeat(int t)
{
	/* Arbitrary, can't find the card's limits */
	if (t < 1 || t > 65535)
		return -EINVAL;

	heartbeat = t;
	wd_heartbeat = t * 100;
	return 0;
}

/**
 *	aio_wdt_get_status:
 *	@status:		the new status.
 *
 *	Extract the status information from a WDT watchdog device. There are
 *	several board variants so we have to know which bits are valid. Some
 *	bits default to one and some to zero in order to be maximally painful.
 *
 *	we then map the bits onto the status ioctl flags.
 */

static int aio_wdt_get_status(int *status)
{
	unsigned char new_status;
	unsigned long flags;

	spin_lock_irqsave(&aio_wdt_lock, flags);
	new_status = inb(WDT_SR);
#ifdef AIO_WDG_DEBUG
        printk(KERN_ERR PFX "Got status: %d\n",new_status);
#endif
	spin_unlock_irqrestore(&aio_wdt_lock, flags);

	*status = 0 || ( new_status & 1);
	if (new_status & WDC_SR_ISOI0)
		*status |= WDIOF_EXTERN1;
	if (new_status & WDC_SR_ISII1)
		*status |= WDIOF_EXTERN2;
	if (type == 501) {
		if (!(new_status & WDC_SR_TGOOD))
			*status |= WDIOF_OVERHEAT;
		if (!(new_status & WDC_SR_PSUOVER))
			*status |= WDIOF_POWEROVER;
		if (!(new_status & WDC_SR_PSUUNDR))
			*status |= WDIOF_POWERUNDER;

		if (tachometer) {
			if (!(new_status & WDC_SR_FANGOOD))
				*status |= WDIOF_FANFAULT;
		}
	}

	return 0;
}

/**
 *	aio_wdt_get_temperature:
 *
 *	Reports the temperature in degrees Fahrenheit. The API is in
 *	farenheit. It was designed by an imperial measurement luddite.
 */

static int aio_wdt_get_temperature(int *temperature)
{
	unsigned short c;
	unsigned long flags;
	spin_lock_irqsave(&aio_wdt_lock, flags);
	c = inb(WDT_RT);
	udelay(8);
	spin_unlock_irqrestore(&aio_wdt_lock, flags);
	*temperature = (c * 11 / 15) + 7;
	return 0;
}

/**
 *	aio_wdt_interrupt:
 *	@irq:		Interrupt number
 *	@dev_id:	Unused as we don't allow multiple devices.
 *
 *	Handle an interrupt from the board. These are raised when the status
 *	map changes in what the board considers an interesting way. That means
 *	a failure condition occurring.
 */

static irqreturn_t aio_wdt_interrupt(int irq, void *dev_id)
{
	/*
	 *	Read the status register see what is up and
	 *	then printk it.
	 */
	unsigned char status;

	spin_lock(&aio_wdt_lock);

	status = inb(WDT_SR);
	udelay(8);

#ifdef AIO_WDG_DEBUG
	printk(KERN_CRIT PFX "status %d\n", status);
#endif
	if (type == 501) {
		if (!(status & WDC_SR_TGOOD)) {
			printk(KERN_CRIT PFX "Overheat alarm.(%d)\n",
								inb(WDT_RT));
			udelay(8);
		}
		if (!(status & WDC_SR_PSUOVER))
			printk(KERN_CRIT PFX "PSU over voltage.\n");
		if (!(status & WDC_SR_PSUUNDR))
			printk(KERN_CRIT PFX "PSU under voltage.\n");

		if (tachometer) {
			if (!(status & WDC_SR_FANGOOD))
				printk(KERN_CRIT PFX "Possible fan fault.\n");
		}
	}
	if (!(status & WDC_SR_WCCR)) {
#ifdef SOFTWARE_REBOOT
#ifdef ONLY_TESTING
		printk(KERN_CRIT PFX "Would Reboot.\n");
#else
		printk(KERN_CRIT PFX "Initiating system reboot.\n");
		emergency_restart(NULL);
#endif
#else
		printk(KERN_CRIT PFX "Reset in 5ms.\n");
#endif
	}
	spin_unlock(&aio_wdt_lock);
	return IRQ_HANDLED;
}


/**
 *	aio_wdt_write:
 *	@file: file handle to the watchdog
 *	@buf: buffer to write (unused as data does not matter here
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 *	A write to a watchdog device is defined as a keepalive signal. Any
 *	write of data will do, as we we don't define content meaning.
 */

static ssize_t aio_wdt_write(struct file *file, const char __user *buf,
						size_t count, loff_t *ppos)
{

	if (count) {
		if (!nowayout) {
			size_t i;

			/* In case it was set long ago */
			expect_close = 0;

			for (i = 0; i != count; i++) {
				char c;
				if (get_user(c, buf + i))
					return -EFAULT;
				if (c == 'V')
					expect_close = 42;
			}
		}
		aio_wdt_ping();
	}
	return count;
}

/**
 *	aio_wdt_ioctl:
 *	@file: file handle to the device
 *	@cmd: watchdog command
 *	@arg: argument pointer
 *
 *	The watchdog API defines a common set of functions for all watchdogs
 *	according to their available features. We only actually usefully support
 *	querying capabilities and current status.
 */
static int aio_wdt_reset(void)
{
#ifdef AIO_WDG_DEBUG
  printk(KERN_ERR PFX "Resetting Watchdog\nStopping counter\n" );
#endif
  aio_wdt_stop();

  aio_wdt_ctr_mode(  0, 3 );
  aio_wdt_ctr_load(  0, 20833 );
  aio_wdt_ctr_mode(  1, 2 );
  aio_wdt_ctr_load(  1, wd_heartbeat );
  aio_wdt_ctr_mode(  2, 1 );

  udelay(100);

  outb(0x90, WDT_COUNT0+3 );

  return 0;
}
/*
 * @arg enable flag for the buzzer
 */
static void aio_wdt_enable_buzzer(int enable)
{
  if( enable )
  	outb( 0x10,WDT_BUZZER );
  else 
	inb( WDT_BUZZER );

}


/**
 *	aio_wdt_ioctl:
 *	@file: file handle to the device
 *	@cmd: watchdog command
 *	@arg: argument pointer
 *
 *	The watchdog API defines a common set of functions for all watchdogs
 *	according to their available features. We only actually usefully support
 *	querying capabilities and current status.
 */

static long aio_wdt_ioctl(struct file *file, unsigned int cmd,
							unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	int new_heartbeat;
	int status;

	static struct watchdog_info ident = {
		.options =		WDIOF_SETTIMEOUT|
					WDIOF_MAGICCLOSE|
					WDIOF_KEEPALIVEPING,
		.firmware_version =	1,
		.identity =		"P104-WDG-CSM/A/E",
	};

	/* Add options according to the card we have */
	ident.options |= (WDIOF_EXTERN1|WDIOF_EXTERN2);
	if (type == 501) {
		ident.options |= (WDIOF_OVERHEAT|WDIOF_POWERUNDER|
							WDIOF_POWEROVER);

		if (tachometer)
		  ident.options |= WDIOF_FANFAULT;
	}
#ifdef AIO_WDG_DEBUG
        printk(KERN_DEBUG PFX "Got the value : %d\n", cmd );
#endif

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		return copy_to_user(argp, &ident, sizeof(ident)) ? -EFAULT : 0;
	case WDIOC_GETSTATUS:
          aio_wdt_get_status(&status);
          return put_user(status, p);
	case WDIOC_GETBOOTSTATUS:
		return put_user(0, p);
	case WDIOC_KEEPALIVE:
		aio_wdt_ping();
		return 0;
	case WDIOC_SETTIMEOUT:
		if (get_user(new_heartbeat, p))
			return -EFAULT;
		if (aio_wdt_set_heartbeat(new_heartbeat))
			return -EINVAL;
		aio_wdt_ping();
		/* Fall */
	case WDIOC_GETTIMEOUT:
		return put_user(heartbeat, p);
        case AIO_WDG_RESET:
          return aio_wdt_reset();
        case AIO_WDG_BUZZER_ENABLE:
          	aio_wdt_enable_buzzer(1);
                return 0;
        case AIO_WDG_BUZZER_DISABLE:
          	aio_wdt_enable_buzzer(0);
                return 0;
	default:
		return -ENOTTY;
	}
}



/**
 *	aio_wdt_open:
 *	@inode: inode of device
 *	@file: file handle to device
 *
 *	The watchdog device has been opened. The watchdog device is single
 *	open and on opening we load the counters. Counter zero is a 100Hz
 *	cascade, into counter 1 which downcounts to reboot. When the counter
 *	triggers counter 2 downcounts the length of the reset pulse which
 *	set set to be as long as possible.
 */

static int aio_wdt_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(0, &open_lock))
		return -EBUSY;

	if (nowayout)
		__module_get(THIS_MODULE);
	/*
	 *	Activate
	 */
	aio_wdt_start();
	return nonseekable_open(inode, file);
}

/**
 *	aio_wdt_release:
 *	@inode: inode to board
 *	@file: file handle to board
 *
 *	The watchdog has a configurable API. There is a religious dispute
 *	between people who want their watchdog to be able to shut down and
 *	those who want to be sure if the watchdog manager dies the machine
 *	reboots. In the former case we disable the counters, in the latter
 *	case you have to open it again very soon.
 */

static int aio_wdt_release(struct inode *inode, struct file *file)
{
	if (expect_close == 42) {
		aio_wdt_stop();
	} else {
		printk(KERN_CRIT PFX "Unexpected close, not stopping timer!");
		aio_wdt_ping();
	}
	expect_close = 0;
	clear_bit(0, &open_lock);
	return 0;
}

/**
 *	aio_wdt_temp_read:
 *	@file: file handle to the watchdog board
 *	@buf: buffer to write 1 byte into
 *	@count: length of buffer
 *	@ptr: offset (no seek allowed)
 *
 *	Read reports the temperature in degrees Fahrenheit. The API is in
 *	fahrenheit. It was designed by an imperial measurement luddite.
 */

static ssize_t aio_wdt_temp_read(struct file *file, char __user *buf,
						size_t count, loff_t *ptr)
{
	int temperature;

	if (aio_wdt_get_temperature(&temperature))
		return -EFAULT;

	if (copy_to_user(buf, &temperature, 1))
		return -EFAULT;

	return 1;
}

/**
 *	aio_wdt_temp_open:
 *	@inode: inode of device
 *	@file: file handle to device
 *
 *	The temperature device has been opened.
 */

static int aio_wdt_temp_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

/**
 *	aio_wdt_temp_release:
 *	@inode: inode to board
 *	@file: file handle to board
 *
 *	The temperature device has been closed.
 */

static int aio_wdt_temp_release(struct inode *inode, struct file *file)
{
	return 0;
}

/**
 *	notify_sys:
 *	@this: our notifier block
 *	@code: the event being reported
 *	@unused: unused
 *
 *	Our notifier is called on system shutdowns. We want to turn the card
 *	off at reboot otherwise the machine will reboot again during memory
 *	test or worse yet during the following fsck. This would suck, in fact
 *	trust me - if it happens it does suck.
 */

static int aio_wdt_notify_sys(struct notifier_block *this, unsigned long code,
							void *unused)
{
	if (code == SYS_DOWN || code == SYS_HALT)
		aio_wdt_stop();
	return NOTIFY_DONE;
}

/*
 *	Kernel Interfaces
 */


static const struct file_operations aio_wdt_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= aio_wdt_write,
	.unlocked_ioctl	= aio_wdt_ioctl,
	.open		= aio_wdt_open,
	.release	= aio_wdt_release,
};

static struct miscdevice aio_wdt_miscdev = {
	.minor	= WATCHDOG_MINOR,
	.name	= "watchdog",
	.fops	= &aio_wdt_fops,
};

static const struct file_operations aio_wdt_temp_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.read		= aio_wdt_temp_read,
	.open		= aio_wdt_temp_open,
	.release	= aio_wdt_temp_release,
};

static struct miscdevice temp_miscdev = {
	.minor	= TEMP_MINOR,
	.name	= "temperature",
	.fops	= &aio_wdt_temp_fops,
};

/*
 *	The WDT card needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off.
 */

static struct notifier_block aio_wdt_notifier = {
	.notifier_call = aio_wdt_notify_sys,
};


static int __devinit aio_wdt_init_one(struct pci_dev *dev, 
				  const struct pci_device_id *ent)
{
	int ret = -EIO;
        printk(KERN_ERR PFX "Initializing PCI Watchdog\n");
        
	dev_count++;
	if (dev_count > 1) {
		printk(KERN_ERR PFX "This driver only supports one device\n");
		return -ENODEV;
	}

	if (type != 500 && type != 501) {
		printk(KERN_ERR PFX "unknown card type '%d'.\n", type);
		return -ENODEV;
	}

	if (pci_enable_device(dev)) {
		printk(KERN_ERR PFX "Not possible to enable PCI Device\n");
		return -ENODEV;
	}

	if (pci_resource_start(dev, 2) == 0x0000) {
		printk(KERN_ERR PFX "No I/O-Address for card detected\n");
		ret = -ENODEV;
		goto out_pci;
	}

	if (pci_request_region(dev, 2, "aio_wdt")) {
		printk(KERN_ERR PFX "I/O address 0x%llx already in use\n",
			(unsigned long long)pci_resource_start(dev, 2));
		goto out_pci;
	}

	irq = dev->irq;
	io = pci_resource_start(dev, 2);

	if (request_irq(irq, aio_wdt_interrupt, IRQF_DISABLED | IRQF_SHARED,
			 "aio_wdt", &aio_wdt_miscdev)) {
		printk(KERN_ERR PFX "IRQ %d is not free\n", irq);
		goto out_reg;
	}

	printk(KERN_INFO
	 "PCI-WDT500/501 (PCI-WDG-CSM) driver 0.10 at 0x%llx (Interrupt %d)\n",
					(unsigned long long)io, irq);

	/* Check that the heartbeat value is within its range;
	   if not reset to the default */
	if (aio_wdt_set_heartbeat(heartbeat)) {
		aio_wdt_set_heartbeat(WD_TIMO);
		printk(KERN_INFO PFX
		  "heartbeat value must be 0 < heartbeat < 65536, using %d\n",
								WD_TIMO);
	}

	ret = register_reboot_notifier(&aio_wdt_notifier);
	if (ret) {
		printk(KERN_ERR PFX
			"cannot register reboot notifier (err=%d)\n", ret);
		goto out_irq;
	}

	if (type == 501) {
		ret = misc_register(&temp_miscdev);
		if (ret) {
			printk(KERN_ERR PFX
			"cannot register miscdev on minor=%d (err=%d)\n",
							TEMP_MINOR, ret);
			goto out_rbt;
		}
	}

	ret = misc_register(&aio_wdt_miscdev);
	if (ret) {
		printk(KERN_ERR PFX
			"cannot register miscdev on minor=%d (err=%d)\n",
						WATCHDOG_MINOR, ret);
		goto out_misc;
	}

        /* aio_wdt_reset_card( dev , ) ; */
        

	printk(KERN_INFO PFX "initialized. heartbeat=%d sec (nowayout=%d)\n",
		heartbeat, nowayout);
	if (type == 501)
		printk(KERN_INFO "wdt: Fan Tachometer is %s\n",
				(tachometer ? "Enabled" : "Disabled"));

	ret = 0;

        aio_wdt_reset();


out:
	return ret;

out_misc:
	if (type == 501)
		misc_deregister(&temp_miscdev);
out_rbt:
	unregister_reboot_notifier(&aio_wdt_notifier);
out_irq:
	free_irq(irq, &aio_wdt_miscdev);
out_reg:
	pci_release_region(dev, 2);
out_pci:
	pci_disable_device(dev);
	goto out;
}


static void __devexit aio_wdt_remove(struct pci_dev *pdev)
{
	/* here we assume only one device will ever have
	 * been picked up and registered by probe function */
	misc_deregister(&aio_wdt_miscdev);
	if (type == 501)
		misc_deregister(&temp_miscdev);
	unregister_reboot_notifier(&aio_wdt_notifier);
	free_irq(irq, &aio_wdt_miscdev);
	pci_release_region(pdev, 2);
	pci_disable_device(pdev);
	dev_count--;
}


static struct pci_device_id aio_wdt_pci_tbl[] = {
	{
		.vendor	   = PCI_VENDOR_ID_ACCESSIO,
		.device	   = PCI_DEVICE_ID_WDG_CSM,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
	},
	{ 0, }, /* terminate list */
};
MODULE_DEVICE_TABLE(pci, aio_wdt_pci_tbl);


static struct pci_driver aio_wdt_driver = {
	.name		= "wdt_pci",
	.id_table	= aio_wdt_pci_tbl,
	.probe		= aio_wdt_init_one,
	.remove		= __devexit_p(aio_wdt_remove),
};


/**
 *	aio_wdt_cleanup:
 *
 *	Unload the watchdog. You cannot do this with any file handles open.
 *	If your watchdog is set to continue ticking on close and you unload
 *	it, well it keeps ticking. We won't get the interrupt but the board
 *	will not touch PC memory so all is fine. You just have to load a new
 *	module in xx seconds or reboot.
 */

static void __exit aio_wdt_cleanup(void)
{
	pci_unregister_driver(&aio_wdt_driver);
}

/**
 * 	aio_wdt_init:
 *
 *	Set up the WDT watchdog board. All we have to do is grab the
 *	resources we require and bitch if anyone beat us to them.
 *	The open() function will actually kick the board off.
 */

static int __init aio_wdt_init(void)
{
	return pci_register_driver(&aio_wdt_driver);
}


module_init(aio_wdt_init);
module_exit(aio_wdt_cleanup);

MODULE_AUTHOR("Jimi Damon ");
MODULE_DESCRIPTION("Driver for the Acces I/O PCIeWDG-CSM/CSMA watchdog cards");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
MODULE_ALIAS_MISCDEV(TEMP_MINOR);
