/*

    This is a modification of a PCI-AI12-16(A) driver to support the LPCI-A16-16A instead.

    - Dare

*/

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *    This is a device driver for the ACCES I/O Products ADC Converter
 *    board, Model PCI-AI12-16.
 *
 *    This is free software, written by Richard B. Johnson. It is
 *    presumed to be correct but it is not warranted to do anything
 *    useful. Therefore, one must use this at their own peril since
 *    it is possible for any kernel code to corrupt a machine and
 *    destroy valuable property existing on storage media.
 *
 *    This software is Copyright(c) 2003, by Richard B. Johnson
 *    Everybody is granted a non-exclusive license to use this
 *    software for any purpose. However, you are required to retain
 *    the author's name within this source-code.
 *
 *    This software is written using publicly-available data, an
 *    instruction manual downloaded from the vendor's Web Page.
 *    Additional information was obtained experimentally. Therefore
 *    it is believed that no proprietary nor trade-secret information
 *    was used to write this driver.
 *
 *    Please send any bug-fixes or improvements to rjohnson@analogic.com
 *
 *    I have implemented two modes of operation.
 *
 *    (1)  The streaming mode where one channel is used to
 *    acquire data as rapidly as possible. The conversion is timed
 *    by the internal timers so that there is a known sample-rate
 *    that can be used for further data processing.
 *
 *    (2)  The multiplexing mode where all channels are converted in
 *    a round-robin fashion. The data from the channels are filtered
 *    using an IIR filter, and the data may be read at any time without
 *    upsetting the conversion process.  The filtered data are read
 *    from memory without affecting the conversion process.
 *
 *    This is the mode one would normally use to acquire noisy data
 *    in an industrial environment.
 *
 *    In both cases, the sample-rate can be set from user-mode via
 *    ioctl(), and, in the case of the multiplexing mode, the IIR
 *    filter coefficient may be set as well.
 *
 *    Note that this will only run on an ix86 platform, not the
 *    other Linux ones like Sparc, Power/PC, etc.
 *
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/wait.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/pci.h>
#include <asm/io.h>
#include "adcboard.h"

#define ADC_CLASS   0xff00
#define ADC_DEVICE  0xECE8
#define ADC_VENDOR  0x494f
#define TRUE 1

#define DEBUG

#ifdef  DEBUG
#define DEB(f) f
#else
#define DEB(f)
#endif

static const char devname[]="ACCES I/O Products LPCI-A16-16A";
static const char vers[]="V1.0";
MODULE_AUTHOR("Richard B. Johnson");
MODULE_DESCRIPTION("Driver for LPCI-A16-16A A/D converter board");
MODULE_SUPPORTED_DEVICE("LPCI-A16-16A");
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Macros. These are supposed to reduce code clutter.
 */
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*
    ...But they generally don't. - Dare
*/

/*
 *   The ADC Control registers.
 */
#define ADC_CLEARFIFO  0x0001
#define ADC_CHAN       0x0002
#define ADC_BURST      0x0003
#define ADC_STAT       0x0008
#define ADC_DSTAT      0x0009
#define ADC_IEN        0x000C
#define ADC_TIM0       0x0014
#define ADC_TIM1       0x0015
#define ADC_TIM2       0x0016
#define ADC_TIMC       0x0017
#define ADC_OVGO       0x001A
#define ADC_CSRC       0x001B
#define ADC_DELAY      0x001D
#define ADC_CGAT       0x001E

#define ADC_DATA       0x0000
#define DAC_0          0x0004

#define ADC_FE         0x80
#define ADC_FF         0x40
#define ADC_FH         0x20
#define ADC_MFF        0x01

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#define BUF_LEN 0x1000
#define BUF_MSK (BUF_LEN -1)
#define DIVISOR 100                 /* Default timer divisor  */
#define CTL_LEN 0x10                /* Length of I/O space    */
#define NR_CHAN 0x10                /* Number of A/D Channels */
#define INTERRUPT_TIMEOUT (1 * HZ)
#define PCI_INDEX(v) (PCI_BASE_ADDRESS_0 + ((v) * sizeof(long)))
#define ArraySize(a) (sizeof(a) / sizeof(a[0]))

typedef struct {
   union {
      struct adc_multi ch;     /* Individual channel data   */
      int val[NR_CHAN];        /* Same, but in an array     */
   } var;                       /* For ADC variables         */
} ADC;

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  This is the device-specific information and data storage. There is
 *  only one global data element, a pointer which points to this structure.
 */

typedef struct {
    int buf[BUF_LEN];         /* ADC Data buffer       */
    int out[BUF_LEN];         /* ADC Data buffer       */
    char dev[0x30];           /* Device name           */
    ADC adc;                  /* ADC data              */
    struct pci_dev *pci;      /* Save to shut down     */
    spinlock_t adc_lock;      /* Spin-lock variable    */
    wait_queue_head_t wait;   /* For poll (select)     */
    int head;                 /* Circular buffer index */
    int tail;                 /* Circular buffer index */
    int major;                /* Device major number   */
    int irq;                  /* Interrupt to use      */
    atomic_t unload;          /* Unload flag           */
    atomic_t busy;            /* Busy flag             */
    size_t lastchan;          /* Last taken multi-channel channel */
    size_t adc_chan;          /* Single input channel in use */
    unsigned int inters;      /* Number of interrupts  */
    unsigned int poll_mask;   /* For select() (poll)   */
    unsigned int base16;      /* PCI/Bus IO address 2  */
    unsigned int base8;       /* PCI/Bus IO address 3  */
    unsigned int acq_mode;    /* Data acquisition mode */
    unsigned int timer;       /* ADC timer divisor     */
    unsigned long foobar;     /* Error code, or 0      */
    } INFO;

static INFO *info;

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Function prototypes.
 */
int __init init_module(void);
void cleanup_module(void);
static int adc_close(struct inode *inp, struct file *fp);
static int adc_open(struct inode *inp, struct file *fp);
static int adc_read(struct file *fp, char *cp, size_t len, loff_t *ppos);
static int adc_ioctl(struct inode *inp, struct file *fp, size_t len, unsigned long ptr);
static unsigned int adc_poll(struct file *fp, struct poll_table_struct *pt);
static void adc_isr(int irq, void *lcl, struct pt_regs *ptreg);
static void acq_data(INFO *inf);
static void card_setup(INFO *inf);
static void set_mode(INFO *inf);
static void stop_board(INFO *inf);
static void ctr_mode(unsigned int addr, unsigned int cntr, unsigned int mode);
static void ctr_load(unsigned int addr, unsigned int cntr, unsigned int val);
static void set_point_list(INFO *inf);
static int remote_set_point_list(INFO *inf, struct adc_list *sdc);
static void clear_all_data(INFO *inf);

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   File operation structure contains function pointers to various routines.
 */
static struct file_operations  adc_fops = {
    owner             : THIS_MODULE ,
    llseek            : NULL ,
    read              : adc_read ,
    write             : NULL ,
    readdir           : NULL ,
    poll              : adc_poll ,
    ioctl             : adc_ioctl ,
    mmap              : NULL ,
    open              : adc_open ,
    flush             : NULL ,
    release           : adc_close ,
    fsync             : NULL ,
    fasync            : NULL ,
    lock              : NULL ,
    readv             : NULL ,
    writev            : NULL ,
    sendpage          : NULL ,
    get_unmapped_area : NULL
};

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  clear all the channel data at once.
 */
static void clear_all_data(INFO *inf)
{
    memset(inf->adc.var.val, 0x00, sizeof(inf->adc.var.val));
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Set a counter mode.
 */
static void ctr_mode(unsigned int addr, unsigned int cntr, unsigned int mode)
{
    outb((cntr << 6) | 0x30 | (mode << 1), addr + ADC_TIMC);
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Load a counter divisor.
 */
static void ctr_load(unsigned int addr, unsigned int cntr, unsigned int val)
{
    outb( val & 0xff         , addr + cntr + ADC_TIM0);
    outb((val >> 0x08) & 0xff, addr + cntr + ADC_TIM0);
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Acquire the data. Note that this is called under a spin-lock.
 */
#define FOR(_REPS_, _CONTENT_) { int I; for ( I = 0; I < _REPS_; ++I ) _CONTENT_; }

static void acq_data(INFO *inf)
{
    int new, sav, i;
    size_t chan;
    unsigned char Status;

    if ( (inb(inf->base8 + ADC_DSTAT) & ADC_MFF) != 0 )
    {
        //The FIFO filled, we're out of sync, discard and start over

        outb(0x00, inf->base8 + ADC_OVGO);            /* Disable all acquisition*/
        outb(0x00, inf->base8 + ADC_CSRC);            /* Disable Ctr trig.      */
        //FOR(32, (void)inb(inf->base8 + ADC_STAT));    /* Waste a bit of time    */
        outb(0x00, inf->base8 + ADC_CLEARFIFO);       /* Reset(empty) FIFO      */
        outb(0x01, inf->base8 + ADC_CSRC);            /* Select Ctr1&2 as trig. */
        outb(0x11, inf->base8 + ADC_OVGO);            /* Enable acquisition     */
        inf->lastchan = -1;
        inf->foobar = DATA_ERROR_FIFO_FILL;
    }
    else
    {
        switch(inf->acq_mode)
        {
        case ADC_SINGLE_CHAN:
            do {
                new = inw(inf->base16 + ADC_DATA);/* Get the converter data   */
                sav = inf->head;                  /* Last buffer position     */
                inf->buf[sav++] = new;            /* Put data into the buffer */
                sav &= BUF_MSK;                   /* Auto-wrap around buffer  */
                if(sav != inf->tail)              /* If room in buffer        */
                    inf->head = sav;              /* Is room in the buffer    */
            } while ( (inb(inf->base8 + ADC_STAT) & ADC_FE) == 0 );
            break;
        case ADC_MULTI_CHAN:
            Status = inb(inf->base8 + ADC_STAT);
            if ( (Status & ADC_FH) == 0 )
            {
                inf->foobar = DATA_ERROR_NO_HALF;
                break;
            }
	    
          for (i=0;i<512 /*HALF_FIFO*/;i++)
          {
              new = inw(inf->base16+ADC_DATA); /* Get the converter data   */
              chan = inf->lastchan = (inf->lastchan + 1) & 0xF;
              inf->adc.var.val[chan] = new;
          }

            while ( -1 )
            {
                Status = inb(inf->base8 + ADC_STAT);
                if ( (Status & ADC_FE) != 0 ) break;

                new = inw(inf->base16+ADC_DATA); // Get the converter data
                chan =
                inf->lastchan = (inf->lastchan + 1) & 0xF;
                inf->adc.var.val[chan] = new;
            }

            break;
        default:
            printk(KERN_ALERT"%s : Unknown acquisition mode (%d)\n",
                                  inf->dev, inf->acq_mode);
            break;
        }
    }

    //outb(0x10, inf->base8 + ADC_IEN);             /* Re-enable FIFO half IRQ */
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This sets up the point-list for the FIFO before acquisition.
 */
static void set_point_list(INFO *inf)
{
    outb(0x00, inf->base8 + ADC_CLEARFIFO);   /* Reset (empty) FIFO */
    if(inf->acq_mode == ADC_SINGLE_CHAN)
        outb(0x11 * inf->adc_chan, inf->base8 + ADC_CHAN);
    else
        outb(0xF0, inf->base8 + ADC_CHAN);
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This interrupts acquisition to set up the point-list for the FIFO, and
 *   also receives a list of gains, which are currently ignored.
 */
static int remote_set_point_list(INFO *inf, struct adc_list *adc)
{
    size_t flags;

    outb(0x00, inf->base8 + ADC_CSRC);            /* Disable Ctr trig.     */
    outb(0x00, inf->base8 + ADC_IEN);             /* Disable FIFO half IRQ */
    spin_lock_irqsave(&inf->adc_lock, flags);

    outb(0x00, inf->base8 + ADC_CLEARFIFO);       /* Reset (empty) FIFO    */
    if(inf->acq_mode == ADC_SINGLE_CHAN)
        outb(0x11 * inf->adc_chan, inf->base8 + ADC_CHAN);
    else
        outb(0xF0, inf->base8 + ADC_CHAN);

    outb(0xC0, inf->base8 + ADC_CGAT);            /* Enable counter gates   */
    outb(0x01, inf->base8 + ADC_CSRC);            /* Select Ctr1&2 as trig. */
    outb(0x10, inf->base8 + ADC_IEN);             /* Enable FIFO half IRQ   */
    inf->head = inf->tail = 0;                    /* Clear interrupt buffer */
    clear_all_data(inf);                          /* Clear data array       */
    inf->poll_mask = 0;                           /* Clear poll mask        */
    wake_up_interruptible(&inf->wait);            /* No data are avaiable   */

    spin_unlock_irqrestore(&inf->adc_lock, flags);
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Do various one-time setup steps for the card. This is not called under
 *   a lock. Therefore, the board must be stopped first before calling
 *   this.
 */
static void card_setup(INFO *inf)
{
    outb(0x00, inf->base8 + ADC_BURST); /* Disable burst mode   */
    outb(0x10, inf->base8 + ADC_DELAY); /* Disable delayed mode */
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Set the conversion mode for the device. This is not called under
 *   a lock. Therefore, the board must be stopped first before calling
 *   this. This will restart the board.
 */
static void set_mode(INFO *inf)
{
    int count;
    size_t flags;
    stop_board(inf);
    count = 2;
    while((inf->timer/count) > 65534)          /* Last even count  */
        count <<=1;
    count |= (inf->timer & 1);                 /* In case it's odd */
    DEB(printk("Timer value = %d\n", inf->timer));
    DEB(printk("Count value = %d\n", count));

    spin_lock_irqsave(&inf->adc_lock, flags);
    switch(inf->acq_mode)
    {
    case ADC_SINGLE_CHAN:          /* Acquire streaming data from chan 0 */
        ctr_mode(inf->base8, 0, 2);       /* Counter 0 to square-wave (2) */
        ctr_mode(inf->base8, 1, 2);       /* Counter 1 to square-wave (2) */
        ctr_mode(inf->base8, 2, 2);       /* Counter 2 to square-wave (2) */
        //ctr_load(inf->base8, 0, 0xffff);           /* Counter 0 divisor   */
        ctr_load(inf->base8, 1, count);            /* Counter 1 divisor   */
        ctr_load(inf->base8, 2, inf->timer/count); /* Counter 2 divisor   */
        set_point_list(inf);
        outb(0x11, inf->base8 + ADC_OVGO);
        break;
    case ADC_MULTI_CHAN:           /* Acquire data from all channels    */
        ctr_mode(inf->base8, 0, 2);       /* Counter 0 to square-wave (2) */
        ctr_mode(inf->base8, 1, 2);       /* Counter 1 to square-wave (2) */
        ctr_mode(inf->base8, 2, 2);       /* Counter 2 to square-wave (2) */
        //ctr_load(inf->base8, 0, 0xffff);           /* Counter 0 divisor   */
        ctr_load(inf->base8, 1, count);            /* Counter 1 divisor   */
        ctr_load(inf->base8, 2, inf->timer/count); /* Counter 2 divisor   */
        set_point_list(inf);
        outb(0x11, inf->base8 + ADC_OVGO);
        break;
    default:
        printk(KERN_ALERT"%s : Unknown acquisition mode (%d)\n",
                              inf->dev, inf->acq_mode);
        break;
    }
    inf->head = inf->tail = 0;                  /* Clear interrupt buffer */
    clear_all_data(inf);                        /* Clear data array       */
    inf->poll_mask = 0;                         /* Clear poll mask        */
    wake_up_interruptible(&inf->wait);          /* No data are avaiable   */
    spin_unlock_irqrestore(&inf->adc_lock, flags);
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   This stops the board from converting and interrupting.
 */
static void stop_board(INFO *inf)
{
    outb(0x00, inf->base8 + ADC_IEN);             /* Disable FIFO half IRQ  */
    outb(0x00, inf->base8 + ADC_CGAT);            /* Disable counter gates  */
    outb(0x00, inf->base8 + ADC_CSRC);            /* Select no trig.        */

    inf->head = inf->tail = 0;             /* Clear interrupt buffer  */
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Device poll. We execute a wake-up from the ISR everytime data
 *   are available. Therefore, there is a spin-lock around the poll
 *   mask change.
 */
static unsigned int adc_poll(struct file *fp, struct poll_table_struct *pt)
{
    size_t flags;
    unsigned int mask;
    poll_wait(fp, &info->wait, pt);
    spin_lock_irqsave(&info->adc_lock, flags);
    mask = info->poll_mask;
    info->poll_mask = 0;
    spin_unlock_irqrestore(&info->adc_lock, flags);
    return mask;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  For device control. We set various modes and parameters from here
 *  and also grab a copy of the channel data.
 */
static int adc_ioctl(struct inode *inp, struct file *fp, unsigned int cmd, unsigned long ptr)
{
    size_t i;
    size_t flags;
    int tmp;
    struct adc_list *adc;
    switch(cmd)
    {
    case ADC_MULTI_CHAN:           /* Acquire data from all channels     */
        info->lastchan = -1;
    case ADC_SINGLE_CHAN:          /* Acquire streaming data from chan 0 */
        info->foobar = 0x00000000; /* Clear driver error code            */
        (void) inb(info->base8 + ADC_DSTAT);
        info->acq_mode = cmd;      /* Set data acquisition mode          */
        set_mode(info);            /* Set up the board                   */
        break;

    case ADC_READ_MULTI:           /* Read all data                      */
/*
 *   Copy the data under a lock from the inner buffer to an output buffer,
 *   then copy that buffer to user.
 */
        spin_lock_irqsave(&info->adc_lock, flags);
        for(i=0; i< ArraySize(info->adc.var.val); i++)
        {
            if (info->foobar)
                info->out[i] = info->foobar;
            else
                info->out[i] = info->adc.var.val[i];
        }
        spin_unlock_irqrestore(&info->adc_lock, flags);

        if(copy_to_user((char *)ptr, info->out, sizeof(struct adc_multi)))
            return -EFAULT;
        break;

    case ADC_SET_FILTER:
        printk(KERN_INFO"%s: Filter set command eaten; not supported.\n", info->dev);
        break;

    case ADC_SET_TIMER:
        if(get_user(tmp, (int *)ptr))
            return -EFAULT;
        if(tmp <= 0)
            return -EINVAL;
        info->timer = tmp;
        set_mode(info);              /* Restart with new parameters      */
        break;

    case ADC_SET_CHAN:               /* Set a new input channel */
        if(get_user(tmp, (int *)ptr))
            return -EFAULT;
        if((tmp < 0) || (tmp > 0x0F))
            return -EINVAL;
        info->adc_chan = tmp;
        set_mode(info);              /* Restart with new parameters      */
        break;

    case ADC_DIAGS:
        info->adc_chan  = 0;
        info->acq_mode  = ADC_SINGLE_CHAN; /* Data acquisition mode      */
        info->timer     = 100;             /* Small counter value        */
        set_mode(info);
        for(i=0; i< 0x10; i++)
            ((char *)info->out)[i] = (char)inb(info->base8 + i);
        if(copy_to_user((char *)ptr, info->out, 0x10))
            return -EFAULT;
        break;

    case ADC_POINT_LIST:
        adc = (struct adc_list *) info->out; 
        if(copy_from_user(info->out, (char *)ptr, sizeof(struct adc_list)))
            return -EFAULT; 
        return remote_set_point_list(info, adc);

    case ADC_GET_INTRS:
        if(put_user(info->inters, (int *)ptr))
            return -EFAULT;
        break;

    case ADC_GET_ADDR:
        info->out[0] = info->base16;
        info->out[1] = info->base8;
        if( copy_to_user((char *)ptr, info->out, 8) )
            return -EFAULT;
        break;

    case ADC_SET_DAC0:
        if(get_user(tmp, (int *)ptr))
            return -EFAULT;
        outb(tmp, info->base16 + DAC_0);
        break;

    /* PCI-AI12-16(A) sets mode via software, LPCI-A16-16A sets mode via jumpers */
    case ADC_BIP_10:
    case ADC_BIP_05:
    case ADC_BIP_2d5:
    case ADC_BIP_1d25:
    case ADC_UNI_10:
    case ADC_UNI_05:
    case ADC_UNI_1d25:
    case ADC_UNI_6d25:
        printk(KERN_INFO"%s: Range set command eaten; not supported.\n", info->dev);
        /* return -EINVAL; */
        break;

    default:
        return -EINVAL;
    }
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Read data from the device. Note that the input and output lengths
 *   are in bytes, although the data are in 4-byte integers. The byte-
 *   lengths are in keeping with the de facto Unix standard so sizeof()
 *   can be used from the API.
 */
#if 1
static int adc_read(struct file *fp, char *cp, size_t len, loff_t *ppos)
{
    size_t flags;
    ssize_t i = 0;
    len /= sizeof(int);                              /* Make word-length */
    spin_lock_irqsave(&info->adc_lock, flags);
    while((info->head != (info->tail &= BUF_MSK)) && (len--))
        info->out[i++] = (unsigned short)(info->buf[info->tail++]);
    spin_unlock_irqrestore(&info->adc_lock, flags);
    if((i *= sizeof(int) ))                          /* Make byte-count  */
    {
        if(copy_to_user(cp, info->out, i))
            return -EFAULT;
    }
    return i;
}
#else
static int adc_read(struct file *fp, char *cp, size_t len, loff_t *ppos)
{
    size_t flags;
    ssize_t i = 0;
    len /= sizeof(int);                              /* Make word-length */
    spin_lock_irqsave(&info->adc_lock, flags);
    while((info->head != (info->tail &= BUF_MSK)) && (len--))
    {
        info->out[i++] = i;
        info->tail++;
    }
    spin_unlock_irqrestore(&info->adc_lock, flags);
    if((i *= sizeof(int) ))                          /* Make byte-count  */
    {
        if(copy_to_user(cp, info->out, i))
            return -EFAULT;
    }
    return i;
}
#endif
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Open the device.
 */
static int adc_open(struct inode *inp, struct file *fp)
{
    atomic_inc(&info->busy);                      /* Show it's busy now   */
    if(atomic_read(&info->unload))                /* Is it being unloaded */
    {
        atomic_dec(&info->busy);                  /* Bump the busy flag   */
        return -ENODEV;                           /* It's being unloaded  */
    }
    MOD_INC_USE_COUNT;                            /* Bump usage count     */
    if(atomic_read(&info->busy) == 1)             /* First user open      */
        set_mode(info);
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Close the device.
 */
static int adc_close(struct inode *inp, struct file *fp)
{
    MOD_DEC_USE_COUNT;                         /* Decrement usage count */
    atomic_dec(&info->busy);
    if(!atomic_read(&info->busy))              /* If the last user      */
        stop_board(info);                      /* Stop interrupts       */
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Simple ADC interrupt service routine.
 */
static void adc_isr(int irq, void *lcl, struct pt_regs *ptreg)
{
    spin_lock(&info->adc_lock);
    acq_data(info);                          /* Get the data             */
    info->poll_mask = POLLIN|POLLRDNORM;     /* Set poll mask            */
    spin_unlock(&info->adc_lock);
    info->inters++;                          /* Show we got an interrupt */
    wake_up_interruptible(&info->wait);      /* Show data are avaiable   */
    outb(0x10, info->base8 + ADC_IEN);        /* Re-enable FIFO half IRQ  */
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Initialize and register the module. The device is on a PCI bridge.
 *  FYI, there are two `gotos` in this code. If goto freaks send me
 *  patches to remove them, they are wasting their time.
 */
int __init init_module()
{
    int result;
    struct pci_dev *pdev = NULL;
    unsigned int ioport2, ioport3;
    unsigned long timer;

    while ((pdev = pci_find_class(ADC_CLASS << 8, pdev)))
    {
        DEB(printk("Found %08x (%08x)\n", pdev->vendor, pdev->device));
        if( (pdev->vendor == ADC_VENDOR) &&
            (pdev->device == ADC_DEVICE) )
            goto dev_found;
    }
    printk(KERN_INFO"%s: PCI device not found\n", devname);
    return -ENODEV;
    dev_found: pci_disable_device(pdev);          /* Disable for now      */

    pci_read_config_dword(pdev, PCI_INDEX(2), &ioport2);
    pci_read_config_dword(pdev, PCI_INDEX(3), &ioport3);
    ioport2 ^= 0x0001;
    ioport3 ^= 0x0001;
    if ( (ioport2 | ioport3) & 0x0001 )
    {
        printk(KERN_INFO"%s: I/O space not found for ADC board\n", devname);
        return -ENODEV;
    }

    (void)pci_enable_device(pdev); /* Now enable the device  */
    if((info = (INFO *) kmalloc(sizeof(INFO), GFP_KERNEL)) == NULL)
    {
        printk(KERN_CRIT"%s: Can't allocate memory\n", devname);
        pci_disable_device(pdev);             /* Disable the device     */
        return -ENOMEM;
    }
    memset(info, 0x00, sizeof(INFO));         /* Clear the structure  */
    strcpy(info->dev, devname);
    info->pci      = pdev;
    info->major    = DEV_MAJOR;
    info->adc_lock = SPIN_LOCK_UNLOCKED;
    info->irq      = pdev->irq;
    info->base16   = ioport2;
    info->base8    = ioport3;
    info->timer    = DIVISOR;
    info->acq_mode = ADC_MULTI_CHAN;
    info->foobar   = 0x00000000;
    atomic_inc(&info->busy);
    atomic_inc(&info->unload);
    stop_board(info);                          /* No interrupts for now */
    init_waitqueue_head(&info->wait);          /* Set up wait queue     */
    if((result = register_chrdev(info->major, info->dev, &adc_fops)) < 0)
    {
        printk(KERN_ALERT"%s: Can't register major number %d\n",
                                              devname, info->major);
        pci_disable_device(pdev);             /* Disable the device     */
        kfree(info);
        return result;
    }
    request_region(info->base8, 64, info->dev);
    request_region(info->base16, 64, info->dev);
    if((result=request_irq(info->irq, adc_isr, SA_SHIRQ, info->dev, info)) < 0)
    {
        printk(KERN_ALERT"%s: Can't allocate IRQ %d\n", devname, info->irq);
        pci_disable_device(pdev);             /* Disable the device     */
        (void)unregister_chrdev(info->major, info->dev);
        release_region(info->base8, 64);
        release_region(info->base16, 64);
        kfree(info);
        return result;
    }
    card_setup(info);                    /* Do one-time setup for the board */
    set_mode(info);                      /* Start the board */
/*
 *  Let the auto multi-channel mode run for a bit to see if
 *  the board interrupts okay.
 */
    timer = jiffies + INTERRUPT_TIMEOUT;
    while(time_before(jiffies, timer))
    {
        if(info->inters) break;
        schedule_timeout(2);
    }
    stop_board(info);                  /* Stop the board  */
    printk(KERN_INFO"%s: Version    = %s\n",   info->dev, vers);
    printk(KERN_INFO"%s: I/O port16 = %04X\n", info->dev, info->base16);
    printk(KERN_INFO"%s: I/O port8  = %04X\n", info->dev, info->base8);
    printk(KERN_INFO"%s: Interrupt  = %d\n",   info->dev, info->irq);
    printk(KERN_INFO"%s: Dev. major = %d\n",   info->dev, info->major);
    printk(KERN_INFO"%s: Patches to rjohnson@analogic.com\n", info->dev);
    if(!info->inters)
       printk(KERN_ALERT"%s: Warning, no interrupt detected\n", info->dev);
    printk(KERN_INFO"%s: Initialization complete\n", info->dev);
    atomic_dec(&info->busy);
    atomic_dec(&info->unload);
    return 0;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *   Release the module.
 */
void cleanup_module()
{
    int result;
    atomic_inc(&info->unload);
    if(atomic_read(&info->busy))
    {
        atomic_dec(&info->unload);
        printk(KERN_ALERT "%s: Can't unregister, busy.\n", info->dev);
        return;
    }
    stop_board(info);                         /* Stop the board         */
    pci_disable_device(info->pci);            /* Disable the device     */
    if((result = unregister_chrdev(info->major, info->dev)) < 0)
    {
        printk(KERN_ALERT "%s: Can't unregister (%d)\n", info->dev, result);
        return;
    }
    free_irq(info->irq, info);
    release_region(info->base8, 64);
    release_region(info->base16, 64);
    printk(KERN_INFO"%s: Module removed\n", info->dev);
    kfree(info);
    return;
}
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
